-- Analog to Digital
-- Allows you to bind joystick axes to commands that normally require digital
-- controls (e.g. +Shoot1, +Turbo, etc.)
declare('a2d', a2d or {})
a2d.version = 1.2

a2d.color = {}
a2d.color.main = '\127FFFFFF'
a2d.color.reset = '\127o'


--save data
function a2d.save()
	local data = {}
	if a2d.list then
		for js,axes in pairs(a2d.list) do
			data[js] = {}
			if axes then
				for axis,item in pairs(axes) do
					item:save()
					data[js][axis] = 1
				end
			end
		end
	end

	gkini.WriteString('a2d', 'axes', spickle(data))
end


--load data
function a2d.load()
	local data = unspickle(gkini.ReadString('a2d', 'axes', ''))

	for js,axes in pairs(data) do
		for axis,v in pairs(axes) do
			--create a minimal version
			a2d.set(js, axis, 1001, '', -1001, '', true)
			--have it load its data
			a2d.list[js][axis]:load()
		end
	end
end


--print all bindings
function a2d.print()
	if a2d.list then
		for js,axes in pairs(a2d.list) do
			if axes then
				for axis,item in pairs(axes) do
					item:print()
				end
			end
		end
	end
end


--clear all bindings
function a2d.clear_all()
	if a2d.list then
		for js,axes in pairs(a2d.list) do
			if axes then
				for axis,item in pairs(axes) do
					item:unset_bindings()
					a2d.list[js][axis] = nil
				end
			end
		end
	end

	a2d.save()
	print(a2d.color.main .. "All a2d bindings cleared" .. a2d.color.reset)
end


--clear bindings for axis 'axis' on joystick 'js'
function a2d.clear(js, axis)
	if a2d.list and a2d.list[js] and a2d.list[js][axis] then
		a2d.list[js][axis]:unset_bindings()
		a2d.list[js][axis] = nil
	end

	a2d.save()
	print(a2d.color.main .. "a2d bindings for " .. js .. ":" .. axis .." cleared" .. a2d.color.reset)
end


--sets new bindings for axis 'axis' on joystick 'js'
--runs runs '/high_command' when 'high_thresh' is passed in an upwards direction
--runs runs '/low_command' when 'low_thresh' is passed in a downwards direction
--if a command starts with a +, it will run '/command 0' on the way back toward neutral
--exception: for +zoom, it will do -zoom instead of +zoom 0.
--if no_save is set to true, don't save after setting bindings
function a2d.set(js, axis, high_thresh, high_command, low_thresh, low_command, no_save)
	if a2d.list == nil then a2d.list = {} end
	if a2d.list[js] == nil then a2d.list[js] = {} end
	
	a2d.list[js][axis] = a2d.new_item(js, axis, high_thresh, high_command, low_thresh, low_command)

	if not no_save then
		a2d.save()
	end
end


--returns  new bindings for axis 'axis' on joystick 'js'
function a2d.new_item(js, axis, high_thresh, high_command, low_thresh, low_command)
	local item = {}

	item.last = 0

	item.js = js
	item.axis = axis

	--js value is integer in range [-1000, +1000]
	item.high_thresh = high_thresh --values >= this will trigger OnHigh
	item.low_thresh = low_thresh --values <= this will trigger OnLow

	item.high_command = string.lower(high_command)
	item.low_command = string.lower(low_command)

	
	--run the commands
	function item:OnHigh()
		if self.high_command ~= '' then
			gkinterface.GKProcessCommand(self.high_command)
		end
	end
	function item:OffHigh()
		if self.high_command ~= '' and string.sub(self.high_command, 1, 1) == '+' then
			if   self.high_command == '+shoot1'
			  or self.high_command == '+shoot2'
			  or self.high_command == '+shoot3'
			  or self.high_command == '+turbo'
			  then
				gkinterface.GKProcessCommand(self.high_command .. " 0")
			else
				gkinterface.GKProcessCommand('-'..string.sub(self.high_command, 2))
			end
		end
	end
	function item:OnLow()
		if self.low_command ~= '' then
			gkinterface.GKProcessCommand(self.low_command)
		end
	end
	function item:OffLow()
		if self.low_command ~= '' and string.sub(self.low_command, 1, 1) == '+' then
			if   self.low_command == '+shoot1'
			  or self.low_command == '+shoot2'
			  or self.low_command == '+shoot3'
			  or self.low_command == '+turbo'
			  then
				gkinterface.GKProcessCommand(self.low_command .. " 0")
			else
				gkinterface.GKProcessCommand('-'..string.sub(self.low_command, 2))
			end
		end
	end


	--process the motion of the joystick and run any appropriate commands
	function item:process()
		local raw = joystick.GetJoystickSingleAxisValue(self.js, self.axis)

		if raw >= self.high_thresh and self.last < self.high_thresh then
			self:OnHigh()
		elseif raw < self.high_thresh and self.last >= self.high_thresh then
			self:OffHigh()
		end
		if raw <= self.low_thresh and self.last > self.low_thresh then
			self:OnLow()
		elseif raw > self.low_thresh and self.last <= self.low_thresh then
			self:OffLow()
		end

		self.last = raw
	end


	--set the bindings
	--TODO:  back up the old bindings!
	function item:set_bindings()
		if self.js ~= nil and self.axis ~= nil then
			self.command = "a2d_js_" .. self.js .. "_axis_" .. self.axis
			RegisterUserCommand(self.command, function() self:process() end)
			gkinterface.BindJoystickCommand(self.js, self.axis, self.command)
		end
	end


	--unset the bindings
	--TODO:  restore the old bindings!
	function item:unset_bindings()
		if self.command ~= nil then
			UnregisterUserCommand(self.command)
		end
	end


	--save data
	function item:save()
		gkini.WriteInt(self.command, 'js', self.js)
		gkini.WriteInt(self.command, 'axis', self.axis)
		gkini.WriteInt(self.command, 'high_thresh', self.high_thresh)
		gkini.WriteString(self.command, 'high_command', self.high_command)
		gkini.WriteInt(self.command, 'low_thresh', self.low_thresh)
		gkini.WriteString(self.command, 'low_command', self.low_command)
	end


	--load data
	function item:load()
		self.js = gkini.ReadInt(self.command, 'js', self.js)
		self.axis = gkini.ReadInt(self.command, 'axis', self.axis)
		self.high_thresh = gkini.ReadInt(self.command, 'high_thresh', self.high_thresh)
		self.high_command = string.lower(gkini.ReadString(self.command, 'high_command', self.high_command))
		self.low_thresh = gkini.ReadInt(self.command, 'low_thresh', self.low_thresh)
		self.low_command = string.lower(gkini.ReadString(self.command, 'low_command', self.low_command))

		--set the bindings
		item:set_bindings()
	end


	--print data
	function item:print()
		local output = "/a2d set"
		output = output .. " " .. self.js
		output = output .. " " .. self.axis
		output = output .. " " .. self.high_thresh
		output = output .. " " .. self.high_command
		output = output .. " " .. self.low_thresh
		output = output .. " " .. self.low_command
		print(a2d.color.main .. output .. a2d.color.reset)
	end


	--go ahead and set the bindings now
	item:set_bindings()

	return item
end






--display usage information
function a2d.help()
	print(a2d.color.main .. "Analog to Digital" .. a2d.color.reset)
	print(a2d.color.main .. "Usage:" .. a2d.color.reset)
	print(a2d.color.main .. "/a2d help" .. a2d.color.reset)
	print(a2d.color.main .. "/a2d clear <js> <axis>" .. a2d.color.reset)
	print(a2d.color.main .. "/a2d clear_all" .. a2d.color.reset)
	print(a2d.color.main .. "/a2d print" .. a2d.color.reset)
	print(a2d.color.main .. "/a2d set <js> <axis> <high threshold> <high command> <low threshold> <low command>" .. a2d.color.reset)
	print(a2d.color.main .. "Example:  /a2d set 0 4 500 +Shoot1 -500 +Shoot2" .. a2d.color.reset)
end


--process any user commands
function a2d.process_commands(_, data)
	if data and data[1] then
		if data[1] == "set" then
			if data[2] and data[3] and data[4] and data[5] and data[6] and data[7] and not data[8] then
				a2d.set(tonumber(data[2]), tonumber(data[3]), tonumber(data[4]), data[5], tonumber(data[6]), data[7])
			else
				print(a2d.color.main .. "Incorrect options.  Example:  /a2d set 0 4 500 +Shoot1 -500 +Shoot2" .. a2d.color.reset)
			end
		elseif data[1] == "clear" then
			if data[2] and data[3] and not data[4] then
				a2d.clear(tonumber(data[2]), tonumber(data[3]))
			else
				print(a2d.color.main .. "Incorrect options.  Example:  /a2d clear 0 4" .. a2d.color.reset)
			end
		elseif data[1] == "clear_all" then
			a2d.clear_all()
		elseif data[1] == "print" then
			a2d.print()
		else
			a2d.help()
		end
	else
		a2d.help()
	end
end


a2d.load()

RegisterUserCommand("a2d", a2d.process_commands)

